/*************************************************************************/
/*                                                                       */
/*  libservices - /etc/services handling library.                        */
/*                                                                       */
/*  Written by Gustavo Niemeyer <gustavo@niemeyer.net>                   */
/*                                                                       */
/*  Copyright (C) 2001 Conectiva, Inc.                                   */
/*                                                                       */
/*  This library is free software; you can redistribute it and/or        */
/*  modify it under the terms of the GNU Library General Public          */
/*  License as published by the Free Software Foundation; either         */
/*  version 2 of the License, or (at your option) any later version.     */
/*                                                                       */
/*  This library is distributed in the hope that it will be useful,      */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    */
/*  Library General Public License for more details.                     */
/*                                                                       */
/*  You should have received a copy of the GNU Library General Public    */
/*  License along with this library; if not, write to the                */
/*  Free Software Foundation, Inc., 59 Temple Place - Suite 330,         */
/*  Boston, MA  02111-1307, USA.                                         */
/*                                                                       */
/*************************************************************************/
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "libservices.h"

#define JUMP_SPACE(x) while (*x == ' ' || *x == '\t') x++
#define JUMP_NON_SPACE(x) while (*x != 0 && *x != '\n' && *x != ' ' && *x != '\t') x++
#define JUMP_DIGITS(x) while (*x >= '0' && *x <= '9') x++
#define JUMP_LINE(x) while (*x != 0 && *x != '\n') x++

struct serv *serv_create_new()
{
  struct serv *serv = (struct serv *) malloc(sizeof(struct serv));
  if (serv) {
    memset(serv, 0, sizeof(struct serv));
    serv->flags = SERV_NEW;
  }
  return serv;
}

struct serv *serv_load(const char *filename)
{
  struct serv *ret = NULL, *serv = NULL;
  FILE *f = fopen(filename?filename:"/etc/services","r");
  if (f) {
    struct stat st;
    if (fstat(fileno(f), &st) == 0) {
      char *buf = (char *) malloc(st.st_size+1);
      char *bufp = buf;
      char *buf_mark;
      int len;
      if (buf) {
        struct serv *serv_new, *serv_last = NULL;
        struct serv_alias *alias_new, *alias_last;
        buf[st.st_size] = 0;
        if (fread(buf, 1, st.st_size, f) == st.st_size) {
          while (*bufp) {
            while (*bufp == '\n') bufp++;
            /* Search for something interesting... */
            JUMP_SPACE(bufp);
            /* Is it a comment or an end of line/file? */
            if (*bufp == '#') JUMP_LINE(bufp);
            if (*bufp == 0 || *bufp == '\n') continue;
            /* Ok. We've got a name. Mark it and go ahead. */
            buf_mark = bufp;
            JUMP_NON_SPACE(bufp);
            /* Build a place to save this entry and initalize it. */
            serv_new = (struct serv *) malloc(sizeof(struct serv));
            memset(serv_new, 0, sizeof(struct serv));
            if (serv_last) {
              /* It's not the first. Append it. */
              serv_last->next = serv_new;
              serv_new->prev = serv_last;
            } else {
              /* It's the first one. */
              serv = serv_new;
            }
            /* We've got a new tail. */
            serv_last = serv_new;
            /* Do we have some memory? */
            if (!serv_new) { serv_destroy(serv); break; }
            len = bufp-buf_mark;
            serv_new->name = (char *) malloc(len+1);
            /* What about now? */
            if (!serv_new->name) { serv_destroy(serv); break; }
            /* Great! Our name is safe. */
            memcpy(serv_new->name, buf_mark, len);
            serv_new->name[len] = 0;
            /* Go ahead and look for something else. */
            JUMP_SPACE(bufp);
            if (*bufp == 0 || *bufp == '\n') continue;
            /* If it's not a port, jump this bugous line. */
            if (*bufp < '0' || *buf > '9') { JUMP_LINE(bufp); continue; }
            /* Good, we've got a port also. */
            buf_mark = bufp;
            JUMP_DIGITS(bufp);
            len = bufp-buf_mark;
            serv_new->port = (char *) malloc(len+1);
            if (!serv_new->port) { serv_destroy(serv); break; }
            memcpy(serv_new->port, buf_mark, len);
            serv_new->port[len] = 0;
            /* Ok... now let's find what protocol we're using. */
            /* We need a bar. Drop the line if we don't find one. */
            if (*bufp != '/') { JUMP_LINE(bufp); continue; }
            bufp++;
            if (*bufp == 0 || *bufp == '\n' || *bufp == ' ' || *buf == '\t') continue;
            /* It seems that we have a nice protocol. Mark the position. */
            buf_mark = bufp;
            /* We can't JUMP_NON_SPACE here because we may have a comment glued. */
            while (*bufp != 0 && *bufp != '\n' && *bufp != ' ' && *bufp != '\t' && *bufp != '#') bufp++;
            /* Go ahead and save the protocol. */
            len = bufp-buf_mark;
            serv_new->proto = (char *) malloc(len+1);
            if (!serv_new->proto) { serv_destroy(serv); break; }
            memcpy(serv_new->proto, buf_mark, len);
            serv_new->proto[len] = 0;
            /* Now we may have aliases or a comment. Let's find what it is. */
            JUMP_SPACE(bufp);
            alias_last = NULL;
            while (*bufp != '#' && *bufp != 0 && *bufp != '\n') {
              /* We have one or more aliases! Mark the position and go ahead. */
              buf_mark = bufp;
              /* We can't JUMP_NON_SPACE here because we may have a comment glued. */
              while (*bufp != 0 && *bufp != '\n' && *bufp != ' ' && *bufp != '\t' && *bufp != '#') bufp++;
              /* Build a place to save this alias and initalize it. */
              alias_new = (struct serv_alias *) malloc(sizeof(struct serv_alias));
              memset(alias_new, 0, sizeof(struct serv_alias));
              if (alias_last) {
                /* It's not the first. Append it. */
                alias_last->next = alias_new;
                alias_new->prev = alias_last;
              } else {
                /* It's the first one. */
                serv_new->aliases = alias_new;
              }
              /* We've got a new tail. */
              alias_last = alias_new;
              /* Save the alias as usual. */
              len = bufp-buf_mark;
              alias_new->alias = (char *) malloc(len+1);
              if (!alias_new->alias) { serv_destroy(serv); break; }
              memcpy(alias_new->alias, buf_mark, len);
              alias_new->alias[len] = 0;
              /* Look for something else. */
              JUMP_SPACE(bufp);
            }
            /* Now, we may have just comments. */
            if (*bufp == '#') {
              bufp++;
              JUMP_SPACE(bufp);
              /* Is there something interesting? */
              if (*bufp != 0 && *bufp != '\n') {
                /* Mark it and go ahead. */
                buf_mark = bufp;
                JUMP_LINE(bufp);
                len = bufp-buf_mark;
                serv_new->comments = (char *) malloc(len+1);
                /* What about now? */
                if (!serv_new->comments) { serv_destroy(serv); break; }
                /* Great! Our name is safe. */
                memcpy(serv_new->comments, buf_mark, len);
                serv_new->comments[len] = 0;
              }
            }
            bufp++;
          }
        }
        free(buf);
      }
    }
    fclose(f);
  }
  return serv;
}

struct buffer {
  char *data;
  int data_size;
  int size;
  int block_size;
  int error;
};

struct buffer *buffer_init(int size, int block_size)
{
  struct buffer *buf = (struct buffer *) malloc(sizeof(struct buffer));
  if (buf) {
    buf->data = (char *) malloc(size);
    buf->size = size;
    buf->data_size = 0;
    buf->block_size = block_size;
    buf->error = 0;
    if (!buf->data) {
      free(buf);
      buf = NULL;
    }
  }
  return buf;
}

void buffer_putc(struct buffer *buf, char c)
{
  if (buf->data_size+1 > buf->size) {
    buf->size += buf->block_size;
    buf->data = (char *) realloc(buf->data, buf->size);
  }
  *(buf->data+buf->data_size++) = c;
}

void buffer_puts(struct buffer *buf, const char *source, int size)
{
  if (buf->data_size+size > buf->size) {
    int request_size = buf->data_size+size;
    buf->size = request_size+(buf->block_size-(request_size%buf->block_size));
    buf->data = (char *) realloc(buf->data, buf->size);
  }
  memcpy(buf->data+buf->data_size, source, size);
  buf->data_size += size;
}

void buffer_destroy(struct buffer *buf)
{
  free(buf->data);
  free(buf);
}

void print_line(const char *s)
{
  const char *newline = s;
  while (*newline != '\n' && *newline != 0) newline++;
  fwrite(s, 1, newline-s, stdout);
  printf("\n");
}

int serv_write(const char *filename, struct serv *serv_list)
{
  struct serv *serv = serv_list;
  int ret = 1;
  FILE *f = fopen(filename?filename:"/etc/services","r");
  if (f) {
    struct stat st;
    if (fstat(fileno(f), &st) == 0) {
      char *buf = (char *) malloc(st.st_size+1);
      char *bufp = buf;
      char *buf_mark;
      int len;
      int total_len;
      if (buf) {
        buf[st.st_size] = 0;
        if (fread(buf, 1, st.st_size, f) == st.st_size) {
          /* Usually 256 bytes will be enough for the changes. */
          struct buffer *bufout = buffer_init(st.st_size+256, 1024);
          if (bufout) {
            struct serv_alias *alias;
            FILE *fout;
            while (*bufp) {
              while (*bufp == '\n') buffer_putc(bufout, *bufp++);
              //printf("-> ");
              //print_line(bufp);
              buf_mark = bufp;
              /* Search for something interesting... */
              JUMP_SPACE(bufp);
              /* Is it a comment or an end of line/file? */
              if (*bufp == '#') JUMP_LINE(bufp);
              if (*bufp == 0 || *bufp == '\n') {
                buffer_puts(bufout, buf_mark, bufp-buf_mark);
                continue;
              }
              
              /* Jump the entry since we will remove or insert our own. */
              JUMP_LINE(bufp);
              
              for (;;) {
                if (serv) {
                  /* Should we remove it? */
                  if (serv->flags & SERV_REMOVED) {
                    /* Yes. Drop the line and go on. */
                    if (*bufp == '\n') bufp++;
                  } else {
                    /* Insert our name. */
                    len = strlen(serv->name);
                    buffer_puts(bufout, serv->name, len);
                    /* Check the name size and insert the necessary tabs. */
                    if (len < 8) {
                      buffer_puts(bufout, "\t\t", 2);
                      total_len = 16;
                    } else {
                      buffer_putc(bufout, '\t');
                      total_len = len+(8-(len%8));
                    }
                    
                    /* Insert the service port and protocol. */
                    len = strlen(serv->port);
                    total_len += len+1;
                    buffer_puts(bufout, serv->port, len);
                    buffer_putc(bufout, '/');
                    len = strlen(serv->proto);
                    total_len += len;
                    buffer_puts(bufout, serv->proto, len);
                    if (serv->aliases || serv->comments) {
                      if (total_len < 24) {
                        buffer_puts(bufout, "\t\t", 2);
                        total_len = 32;
                      } else {
                        buffer_putc(bufout, '\t');
                        total_len = total_len+(8-(total_len%8));
                      }
                    }

                    alias = serv->aliases;
                    if (alias) {
                      len = strlen(alias->alias);
                      total_len += len;
                      buffer_puts(bufout, alias->alias, len);
                      alias = alias->next;
                      while (alias) {
                        len = strlen(alias->alias);
                        total_len += len+1;
                        buffer_putc(bufout, ' ');
                        buffer_puts(bufout, alias->alias, len);
                        alias = alias->next;
                      }
                    }
                    if (serv->comments) {
                      if (total_len < 40) {
                        buffer_puts(bufout, "\t\t", 2);
                        total_len = 48;
                      } else {
                        buffer_putc(bufout, '\t');
                        total_len = total_len+(8-(total_len%8));
                      }
                      buffer_puts(bufout, "# ", 2);
                      buffer_puts(bufout, serv->comments, strlen(serv->comments));
                    }
                  }
                  if (serv->next && serv->flags & SERV_NEW) {
                    serv = serv->next;
                    buffer_putc(bufout, '\n');
                  } else {
                    serv = serv->next;
                    if (!serv || serv->next || !(serv->flags & SERV_NEW))
                      break;
                    buffer_putc(bufout, '\n');
                  }
                }
              }
            }
            fclose(f);
            f = NULL;
            fout = fopen(filename?filename:"/etc/services","w");
            if (fout) {
              if (fwrite(bufout->data, 1, bufout->data_size, fout) == bufout->data_size)
                ret = 0;
              fclose(fout);
            }
            buffer_destroy(bufout);
          }
        }
        free(buf);
      }
    }
    if (f) fclose(f);
  }
  return ret;
}

void serv_destroy(struct serv *serv_list)
{
  struct serv *serv, *serv_next;
  struct serv_alias *alias, *alias_next;
  serv = serv_list;
  while (serv) {
    serv_next = serv->next;
    free(serv->name);
    free(serv->port);
    free(serv->proto);
    free(serv->comments);
    alias = serv->aliases;
    while (alias) {
      alias_next = alias->next;
      free(alias->alias);
      free(alias);
      alias = alias_next;
    }
    free(serv);
    serv = serv_next;
  }
}

void serv_destroy_aliases(struct serv_alias *aliases)
{
  struct serv_alias *alias_next;
  while (aliases) {
    alias_next = aliases->next;
    free(aliases->alias);
    free(aliases);
    aliases = alias_next;
  }
}

int serv_has_alias(const struct serv *serv, const char *alias)
{
  struct serv_alias *alias_cur = serv->aliases;
  for (; alias_cur; alias_cur = alias_cur->next) {
    if (strcmp(alias_cur->alias, alias) == 0) {
      return 1;
    }
  }
  return 0;
}

int serv_match_any_alias(const struct serv_alias *first,
                         const struct serv_alias *second)
{
  const struct serv_alias *tmp = second;
  for (; first; first = first->next)
    for (second = tmp; second; second = second->next)
      if (strcmp(first->alias, second->alias) == 0)
        return 1;
  return 0;
}

int serv_add(struct serv *serv_list, struct serv *serv_new)
{
  struct serv *serv_last = NULL, *serv = serv_list;
  int serv_new_port = atoi(serv_new->port);
  int ret = 1;
  for (; serv; serv = serv->next) {
    if (serv_new_port < atoi(serv->port)) {
      serv_new->next = serv;
      if (serv->prev) {
        serv_new->prev = serv->prev;
        serv->prev->next = serv_new;
      }
      serv->prev = serv_new;
      ret = 0;
      break;
    }
    serv_last = serv;
  }
  if (ret != 0 && serv_last) {
    serv_new->prev = serv_last;
    serv_last->next = serv_new;
    ret = 0;
  }
  return ret;
}

/* vim:sw=2:ts=2:et
*/
